/*-
 * Copyright (C) @BABOLO  2002 Dec 04
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static const char copyright[] = "\
@(#)Copyright (C) @BABOLO  2002 Dec 04\n\
@(#)All rights reserved.\n";
static const char rcsid[] = "$Id: cmd_trans.c,v 1.31 2008/03/30 07:03:11 babolo Exp $";
#endif /* not lint */

#include <sys/types.h>
#include <sys/param.h>
#include <sys/wait.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <sysexits.h>
#include <unistd.h>
#include <stdlib.h>
#include <limits.h>
#include <string.h>
#include <stdio.h>
#include <errno.h>
#include <poll.h>
#include <mife.h>
#include <err.h>
#include <multilar.h>
#include <libpq-fe.h>
#include <babolo/BLINflag.h>
#include "pgoblin.h"
#include "pgob.h"

#define DBSET GLOBALDBSET

int
pgoblin_getjob(pgoblin_exenv *exenv, pgoblin_r *r) {
    int ex = EX_OK;
    pgoblin_job *j;
    size_t i;
    FILE *fi;

    ifBLIN_QV4(exenv->options->flags) fprintf(stderr, "+pgoblin_getjob %d\n", r[PGO_CJOB].r);
    if  (r[PGO_CJOB].g.j) {
        if  (!(fi = funopen( (void*)(IO_OUT->onu)
                           , NULL
                           , (int (*)(void*, const char*, int))mife_writ
                           , NULL
                           , NULL
            ) )            ) {
            ifBLIN_QV1(exenv->options->flags) warn("funopen on %d", r[PGO_CJOB].r);
            ex = EX_OSERR;
        } else {
            for (i = 0; i <  MULAR_NEXT(*r[PGO_CJOB].g.j); i++) {
                j = mular_getix(*r[PGO_CJOB].g.j, i);
                if  (!(j->pid)) {
                    ifBLIN_QV1(exenv->options->flags) warnx("No pid in %d", j->cmdn);
                    errno = EINVAL;
                    ex = EX_SOFTWARE;
                } else if ((j->cmdn >= 0) && (j->status == PGOBLIN_NOSTATUS)) {
                    if  (((ex = waitpid(j->pid, &(j->status), WNOHANG)) == j->pid) || !ex) {
                        ex = EX_OK;
                    } else if (ex > 0) {
                        ifBLIN_QV1(exenv->options->flags) warn("Pid %d<>%d in %d", ex, j->pid, j->cmdn);
                        ex = EX_OSERR;
                    } else if (errno == ECHILD || errno == EDEADLK || errno == EAGAIN) {
                        ex = EX_OK;
                    } else {
                        ifBLIN_QV1(exenv->options->flags) warn("Pid %d in %d", j->pid, j->cmdn);
                        ex = EX_OSERR;
                }   }
                fprintf(fi, "%d	%d	%d	%d\n", i, j->cmdn, j->pid, j->status);
            }
            fprintf(fi, "\\.\n");
            if  (fclose(fi) < 0) {
                ifBLIN_QV1(exenv->options->flags) warn("fclose on %d", r[PGO_CJOB].r);
                ex = EX_OSERR;
    }   }   }
    ifBLIN_QV4(exenv->options->flags) fprintf(stderr, "-pgoblin_getjob %d\n", ex);
    return(ex);
}

int
pgoblin_wait(pgoblin_exenv *exenv, pgoblin_r *r) {
    int ex = EX_OK;
    pgoblin_job *j;
    size_t i;

    ifBLIN_QV4(exenv->options->flags) fprintf(stderr, "+pgoblin_wait %d\n", r[PGO_CJOB].r);
    if  (r[PGO_CJOB].g.j) {
        for (i = 0; i <  MULAR_NEXT(*r[PGO_CJOB].g.j); i++) {
            j = mular_getix(*r[PGO_CJOB].g.j, i);
            if  (!(j->pid)) {
                ifBLIN_QV1(exenv->options->flags) warnx("No pid in %d", j->cmdn);
                errno = EINVAL;
                ex = EX_SOFTWARE;
            } else if ((j->cmdn >= 0) && (j->status == PGOBLIN_NOSTATUS)) {
                if  ((ex = waitpid(j->pid, &(j->status), 0)) < 0) {
                    ifBLIN_QV1(exenv->options->flags) warn("Pid %d in %d", j->pid, j->cmdn);
                    ex = EX_OSERR;
                } else if (ex == j->pid) {
                    ex = EX_OK;
                } else {
                    ifBLIN_QV1(exenv->options->flags) warn("Pid %d<>%d in %d", ex, j->pid, j->cmdn);
                    ex = EX_OSERR;
    }   }   }   }
    ifBLIN_QV4(exenv->options->flags) fprintf(stderr, "-pgoblin_wait %d\n", ex);
    return(ex);
}

int
pgoblin_untrap(pgoblin_exenv *exenv, pgoblin_r *r) {
    exenv->exeflags &= ~PGOBLIN_TRAPDEF;
    return(EX_OK);
}

int
pgoblin_trap(pgoblin_exenv *exenv, pgoblin_r *r) {
    exenv->ctrap = exenv->ppoint + 1;
    exenv->exeflags |= PGOBLIN_TRAPDEF;
    exenv->lexlevel++;
    return(EX_OK);
}

int
pgoblin_call(pgoblin_exenv *exenv, pgoblin_r *r) {
    int ex = EX_UNAVAILABLE;
    return(ex);
}

int
pgoblin_eval(pgoblin_exenv *exenv, pgoblin_r *r) {
    int ex = EX_OK;
    pgoblin_prog *pgm = NULL;

    ifBLIN_QV4(exenv->options->flags) fprintf(stderr, "+pgoblin_eval %d\n", r[PGO_CCTL].r);
    if  ((ex = pgoblin_parser(exenv->options, IO_CTL->text, IO_CTL->length, &pgm))) {
        ifBLIN_QV1(exenv->options->flags) warn("Parser failed in #eval");
    } else {
        if  (exenv->options->flags & PGOBLIN_VERIFY) pgoblin_Vdump(exenv->options, pgm);
        if  ((ex = pgoblin_execute(exenv->options, pgm, exenv, exenv, 0))) {
            ifBLIN_QV1(exenv->options->flags) warn("Executor failed in #eval");
    }   }
    ifBLIN_QV4(exenv->options->flags) fprintf(stderr, "-pgoblin_eval %d\n", ex);
    return(ex);
}

int
pgoblin_close(pgoblin_exenv *exenv, pgoblin_r *r) {
    if  (r[PGO_CCON].r) {
        R_CONN->flags &= ~PGOBLIN_CLOSECONN;
    }
    return(EX_OK);
}

int
pgoblin_clear(pgoblin_exenv *exenv, pgoblin_r *r) {
    return(EX_OK);
}

int
pgoblin_return(pgoblin_exenv *exenv, pgoblin_r *r) {
    int ex = EX_OK;

    ifBLIN_QV4(exenv->options->flags) fprintf(stderr, "+pgoblin_return %d\n", r[PGO_CCON].r);
    if  (exenv->exeflags & PGOBLIN_FORKED) exit(ex);
    exenv->exeflags |= PGOBLIN_END;
    ifBLIN_QV4(exenv->options->flags) fprintf(stderr, "-pgoblin_return %d\n", ex);
    return(ex);
}

int
pgoblin_exit(pgoblin_exenv *exenv, pgoblin_r *r) {
    int ex;
    ifBLIN_QV4(exenv->options->flags) fprintf(stderr, "+pgoblin_exit\n");
    ex = pgoblin_return(exenv, r);
    ifBLIN_QV4(exenv->options->flags) fprintf(stderr, "-pgoblin_exit %d\n", ex);
    exit(ex);
}

int
pgoblin_begin(pgoblin_exenv *exenv, pgoblin_r *r) {
    PGresult *res;

    ifBLIN_QV4(exenv->options->flags) fprintf(stderr, "+pgoblin_begin\n");
    if  (!(R_CONN->intran)) {
        PQEXEC(res, "BEGIN", PGRES_COMMAND_OK, -1, R_CONN->conn);
        DBSET->clear(res);
        R_CONN->intran++;
    }
    ifBLIN_QV4(exenv->options->flags) fprintf( stderr, "-pgoblin_begin %d\n", 0);
    return(0);
}

int
pgoblin_end(pgoblin_exenv *exenv, pgoblin_r *r) {
    PGresult *res;

    ifBLIN_QV4(exenv->options->flags) fprintf(stderr, "+pgoblin_end\n");
    if  ((R_CONN->intran)) {
        PQEXEC(res, "END", PGRES_COMMAND_OK, -1, R_CONN->conn);
        DBSET->clear(res);
        R_CONN->intran--;
    }
    ifBLIN_QV4(exenv->options->flags) fprintf( stderr, "-pgoblin_end %d\n", 0);
    return(0);
}

int
pgoblin_listen(pgoblin_exenv *exenv, pgoblin_r *r) {
    int       sock;
    struct pollfd pfd;
    int timeout, i, e, ex = EX_OK;
    int64_t tout;
    PGnotify *notify;
    struct timeval t;
    union {
        int64_t itime;
        u_int32_t d[2];
    } dbtime;
    typedef union {
        u_int64_t p;
        u_int32_t d[2];
    } pmatsemit;

    ifBLIN_QV4(exenv->options->flags) fprintf( stderr, "+pgoblin_listen %d %d %d =%s~\n"
                                             , r[PGO_CCTL].r, r[PGO_CCON].r, r[PGO_CIN].r
                                             , (char*)(IO_CTL->text)
                                             );
    if  ((sock = PQsocket(R_CONN->conn)) < 0) {
        ex = EX_UNAVAILABLE;
        errno = EINVAL;
        /* XXXX     */
    } else if (  !IO_IN->text
              || (IO_IN->length < sizeof(pmatsemit))
              ) {
        ex = EX_DATAERR;
        errno = EINVAL;
        /* XXXX     */
    } else {
        /* Should be 64 bit time wait to in usec from 1970-01-01 00:00:00 UTC *
         * (( date_part('epoch', t)                                           *
         *  + date_part('second', t)                                          *
         *  - floor(date_part('second', t))                                   *
         *  ) * 1000000                                                       *
         * )::int8                                                            *
         **********************************************************************/
        dbtime.d[1] = ntohl(((pmatsemit*)(IO_IN->text))->d[0]);
        dbtime.d[0] = ntohl(((pmatsemit*)(IO_IN->text))->d[1]);
        bzero(&pfd, sizeof(pfd));
        pfd.fd = sock;
        pfd.events = POLLIN | POLLRDNORM | POLLRDBAND | POLLPRI;
        for (e = 1; e && !ex;) {
            if  (gettimeofday(&t, NULL)) {
                ex = EX_OSERR;
                /* XXXX     */
                break;
            }
            tout = dbtime.itime - ((int64_t)1000000LL * t.tv_sec + t.tv_usec);
            if  (tout <= 0) break;
            else if ((tout / 1000) > exenv->options->maxtmout) {
                timeout = exenv->options->maxtmout;
                dbtime.itime = ((int64_t)1000LL * exenv->options->maxtmout)
                           + ((int64_t)1000000LL * t.tv_sec)
                           + t.tv_usec
                           ;
            } else timeout = tout / 1000;
            ifBLIN_QV4(exenv->options->flags) fprintf( stderr, "LISTEN=%s~ fd %d for %d ms\n"
                                                     , (char*)IO_CTL->text, sock, timeout
                                                     );
            if  ((i = poll(&pfd, 1, timeout)) < 0) {
                ex = EX_OSERR;
                /* XXXX     */
                break;
            }
            PQconsumeInput(R_CONN->conn);
            while ( (notify = PQnotifies(R_CONN->conn))) {
                ifBLIN_QV4(exenv->options->flags)
                    fprintf(stderr, "Received NOTIFY=%s~\n", notify->relname);
                if  (!strcmp(notify->relname, IO_CTL->text)) e = 0;
                DBSET->freemem(notify);
    }   }   }
    ifBLIN_QV4(exenv->options->flags) fprintf( stderr, "-pgoblin_listen %d\n", ex);
    return(ex);
}

#define L0 0 /* \0           */
#define Ll 1 /*   */
#define Lb 2 /*        */
#define Ls 3 /*    */
#define Ld 4 /* #            */
#define Lo 5 /*     */
#define Lx 6
static const u_char *clasnames[] =
{ "L0", "Ll", "Lb", "Ls", "Ld", "Lo", "Lx"};

static const u_char class[256] =
{ L0, Lo, Lo, Lo,  Lo, Lo, Lo, Lo,  Lo, Lb, Ll, Ll,  Lo, Ll, Lo, Lo
, Lo, Lo, Lo, Lo,  Lo, Lo, Lo, Lo,  Lo, Lo, Lo, Lo,  Lo, Lo, Lo, Lo
, Lb, Lo, Lo, Ld,  Lo, Lo, Lo, Lo,  Lo, Lo, Lo, Lo,  Lo, Lo, Lo, Lo
, Ls, Ls, Ls, Ls,  Ls, Ls, Ls, Ls,  Ls, Ls, Lo, Lo,  Lo, Lo, Lo, Lo

, Lo, Ls, Ls, Ls,  Ls, Ls, Ls, Ls,  Ls, Ls, Ls, Ls,  Ls, Ls, Ls, Ls
, Ls, Ls, Ls, Ls,  Ls, Ls, Ls, Ls,  Ls, Ls, Ls, Lo,  Lo, Lo, Lo, Ls
, Lo, Ls, Ls, Ls,  Ls, Ls, Ls, Ls,  Ls, Ls, Ls, Ls,  Ls, Ls, Ls, Ls
, Ls, Ls, Ls, Ls,  Ls, Ls, Ls, Ls,  Ls, Ls, Ls, Lo,  Lo, Lo, Lo, Lo

, Lo, Lo, Lo, Lo,  Lo, Lo, Lo, Lo,  Lo, Lo, Lo, Lo,  Lo, Lo, Lo, Lo
, Lo, Lo, Lo, Lo,  Lo, Lo, Lo, Lo,  Lo, Lo, Lo, Lo,  Lo, Lo, Lo, Lo
, Lo, Lo, Lo, Ls,  Lo, Lo, Lo, Lo,  Lo, Lo, Lo, Lo,  Lo, Lo, Lo, Lo
, Lo, Lo, Lo, Ls,  Lo, Lo, Lo, Lo,  Lo, Lo, Lo, Lo,  Lo, Lo, Lo, Lo

, Ls, Ls, Ls, Ls,  Ls, Ls, Ls, Ls,  Ls, Ls, Ls, Ls,  Ls, Ls, Ls, Ls
, Ls, Ls, Ls, Ls,  Ls, Ls, Ls, Ls,  Ls, Ls, Ls, Ls,  Ls, Ls, Ls, Ls
, Ls, Ls, Ls, Ls,  Ls, Ls, Ls, Ls,  Ls, Ls, Ls, Ls,  Ls, Ls, Ls, Ls
, Ls, Ls, Ls, Ls,  Ls, Ls, Ls, Ls,  Ls, Ls, Ls, Ls,  Ls, Ls, Ls, Ls
};

static const u_char *statenames[] =
{ "bg", "bs", "nq", "tc", "tb", "tp", "xx"};

enum states
{ bg /*                                           */
, bs /*                                                 */
, nq /*     ?                       */
, tc /*                           */
, tb /*        */
, tp /*                                       */
, xx
};
static const u_char *pn[] = 
{ "- ", "- ", "- ", "- ", "- ", "- ", "- ", "- "
, "- ", "- ", "Ci", "Cp", "- ", "- ", "Ce", "- "
, "Co", "Cl", "Cf", "- ", "- ", "Cx", "- ", "- "
};

#define Ci  0x200000 /*     */
#define Cp  0x100000 /*               */
#define Ce  0x020000 /*          */
#define Co  0x008000 /*             */
#define Cl  0x004000 /*                 */
#define Cf  0x002000 /*               */
#define Cx  0x000400 /*                    */
#define C_state 0xFF /*           */

static u_int32_t automa[xx][Lx] =
/* \0 L0  \n Ll  \b Lb  Ls   # Ld   * Lo   *    @          */
{{ Cx|xx,    bg,    bs,       bs,    nq,    bs }/* bg @                   */
,{ Cx|xx,    bg,    bs,       bs,    bs,    bs }/* bs @                   */
,{ Cx|xx,    bg,    bs, Co|Cl|tc,    bs,    bs }/* nq @ 1           */
,{ Cx|xx, Cx|xx, Cf|tb,    Cl|tc, Cx|xx, Cx|xx }/* tc              */
,{ Cx|xx, Cx|xx,    tb, Ci|Cp|tp, Cx|xx, Cx|xx }/* tb          */
,{ Cx|xx, Ce|xx, Ce|xx,    Cp|tp, Ce|xx, Ce|xx }/* tp      */
};

int
pgoblin_style(pgoblin_exenv *exenv, pgoblin_r *r) {
    int ex = EX_OK, lfcount, i;
    enum states state;
    u_int32_t control;
    size_t offset, b = 0, d = 0;
    u_char s, p, *q;
    int cmp = 0, rs = 0;

    ifBLIN_QV4(exenv->options->flags) fprintf( stderr, "+pgoblin_style %d %d %d =%s~\n"
                                             , r[PGO_CCTL].r, r[PGO_CIN].r, r[PGO_CSTY].r
                                             , (char*)(IO_CTL->text)
                                             );
    if  (!r[PGO_CSTY].r) {
        ifBLIN_QV1(exenv->options->flags) warnx("Try to change style 0");
        ERROUT(EX_DATAERR, EINVAL);
    }
    if  (!IO_CTL->text) {
        ifBLIN_QV1(exenv->options->flags) warnx("No pgoblin style name");
        ERROUT(EX_DATAERR, EINVAL);
    }
    if  (!IO_IN->text) {
        ifBLIN_QV1(exenv->options->flags) warnx("No pgoblin style text");
        ERROUT(EX_DATAERR, EINVAL);
    }
    /** IO_TEXT + **/
    ;;  /** IO_MIFE + **/
    ;;  ;;  IO_IN->flags &= ~(PGOBLIN_FREETEXT | PGOBLIN_FREEMIFE); /* XXXX */
    ;;  /** IO_MIFE - **/
    /** IO_TEXT - **/
    ifBLIN_QV5(exenv->options->flags) BLIN_STATEHEADER(stderr, pn)
    q = IO_IN->text;
    lfcount = 1;
    for (state = bg, offset = 0, control = 0; state < xx; offset++, state = control & C_state) {
        if  (((IO_IN->flags & PGOBLIN_BINPARM) || IO_IN->length) && (offset >= IO_IN->length)) {
            p = L0;
            s = 0;
        } else {
            s = q[offset];
            p = class[s];
            if  (s == '\n') lfcount++;
        }
        control = automa[state][p];
        ifBLIN_QV5(exenv->options->flags)
            BLIN_STATEBODY(stderr, pn, rs, clasnames[p], q, offset, b, control, statenames, state, C_state, (q[offset + BLIN_STATEVAR_i]>=' '))
        if  (control & Ci) b = offset;
        if  (control & Cp) rs++;
        if  (control & Ce) {
            for (i = 0; i < pgoblin.styletbsz; i++) {
                if  (pgoblin.styletable[i] && pgoblin.styletable[i]->name) {
                    if  (  !strncmp(&q[b], pgoblin.styletable[i]->name, rs)
                        ) {
                        R_STYLE->path = NULL;
                        R_STYLE->flags = (exenv->options->flags & BLIN_VERMASK) | i;
                        if  ((ex = STYLESET->parser(R_STYLE, &q[b + rs])))
                            ifBLIN_QV1(exenv->options->flags) fprintf(stderr, "style parser fail\n");
                        if  (exenv->options->flags & PGOBLIN_VERIFY) STYLESET->dump(R_STYLE);
                        goto out;
            }   }   }
            ifBLIN_QV1(exenv->options->flags) fprintf(stderr, "Invalid style interpretor name\n");
            ERROUT(EX_DATAERR, EINVAL);
        }
        if  (control & Co) {
            cmp = 0;
            d = 0;
            rs = 0;
        }
        if  (control & Cl) {
            cmp |= (((u_char*)IO_CTL->text)[d++] - q[offset]);
            if  (cmp) control = bs;
        }
        if  (control & Cf) {
            cmp |= ((u_char*)IO_CTL->text)[d++];
            if  (cmp) control = bs;
        }
        if  (control & Cx) {
            ifBLIN_QV1(exenv->options->flags) fprintf(stderr, "Erroneous style\n");
            ERROUT(EX_DATAERR, EINVAL);
    }   }
out:
    ifBLIN_QV4(exenv->options->flags) fprintf( stderr, "-pgoblin_style %d\n", ex);
    return(ex);
}
